package so.libcrypt.testapp.testapp;

import android.app.Activity;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothManager;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.Button;
import android.widget.Toast;

import org.apache.commons.math3.analysis.MultivariateFunction;
import org.apache.commons.math3.optimization.PointValuePair;
import org.apache.commons.math3.optimization.direct.NelderMeadSimplex;
import org.apache.commons.math3.optimization.direct.SimplexOptimizer;
import org.apache.commons.math3.util.FastMath;
import org.apache.commons.math3.optimization.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.List;
import java.lang.Math;

public class Empty extends Activity {
    public static final String LOGTAG = "31c3";
    private final static int REQUEST_ENABLE_BT = 1;
    // Initializes Bluetooth adapter.
    public BluetoothAdapter mBluetoothAdapter;

    private Hashtable<String, Advertiser> ADVERTISERS = new Hashtable<String, Advertiser>();

    private ArrayList<Double> s;
    private ArrayList<Float> beacon_variance;
    private ArrayList<float[]> beacon_positions;
    private float[] sx;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_empty);

        if (!getPackageManager().hasSystemFeature(PackageManager.FEATURE_BLUETOOTH_LE)) {
            Toast.makeText(this, "BLE not supported", Toast.LENGTH_SHORT).show();
            finish();
        }

        final BluetoothManager bluetoothManager = (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        mBluetoothAdapter = bluetoothManager.getAdapter();

        // Ensures Bluetooth is available on the device and it is enabled. If not,
        // displays a dialog requesting user permission to enable Bluetooth.
        if (mBluetoothAdapter == null || !mBluetoothAdapter.isEnabled()) {
            Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
            startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT);
        }

        Button button = (Button) findViewById(R.id.calc);
        button.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
                get_position();
            }
        });


        mBluetoothAdapter.startLeScan(new BluetoothAdapter.LeScanCallback() {
            @Override
            public void onLeScan(final BluetoothDevice device, final int rssi, final byte[] scanRecord) {
                runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
//                        String record = "";
//                        for (int i=0; i < scanRecord.length; i++) {
//                            record += String.format("0x%02X ",scanRecord[i]);
//                        }
                        if (!ADVERTISERS.containsKey(device.getAddress())) {
                            float[] pos = new float[3];
                            ADVERTISERS.put(device.getAddress(), new Advertiser(device.getAddress(), pos, new Integer(scanRecord[29]).intValue()));
                        }

                        ADVERTISERS.get(device.getAddress()).add_measurement(rssi);


                        List<String> keys = new ArrayList(ADVERTISERS.keySet());
                        Collections.sort(keys);
                        for (String e : keys) {
                            Log.d(LOGTAG, ADVERTISERS.get(e).toString());
                        }
                    }
                });
            }
        });

            //    setBeaconLayout("m:0-3=4c000215,i:4-19,i:20-21,i:22-23,p:24-24"));
    }

    /**
     * Class get called from nelder-mead
     */
    private static class myfunc extends Empty implements MultivariateFunction {


        /**
         * This should be error-function as of python-code
         * @param x
         * @return
         */
        public double value(double[] x) {
            // x - beacon_positions
            ArrayList<float[]> advers = super.getBeacon_positions();
            List<float[]> keys = new ArrayList(advers);
            ArrayList<float[]> tmp = new ArrayList<float[]>();

            // Python: (x-beacon_positions)**2
            for (float[] element: keys) {
                double new_x = Math.pow(x[0] - element[0], 2);
                double new_y = Math.pow(x[1] - element[1], 2);
                double new_z = Math.pow(x[2] - element[2], 2);
                tmp.add(new float[]{(float) new_x, (float) new_y, (float) new_z});



            }
        }
    }


    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.empty, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        // Handle action bar item clicks here. The action bar will
        // automatically handle clicks on the Home/Up button, so long
        // as you specify a parent activity in AndroidManifest.xml.
        // as you specify a parent activity in AndroidManifest.xml.
        int id = item.getItemId();
        if (id == R.id.action_settings) {
            return true;
        }
        return super.onOptionsItemSelected(item);
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
    }

    public float[] vec_sub(float[] a, float[] b) {
        float[] c = new float[0];
        //TODO Exception wenn nicht beide Arrays 3 Elemente
        for (int i = 0; i <= a.length; i++) {
            c[i] = a[i] - b[i];
        }

        return c;
    }

    /**
     * Find Advertiser with lowest distance
     *
     * @return Advertiser with lowest distance
     */
    private Advertiser advertiser_min_distance() {
        // Build a very far Advertiser to be sure to not be the one with the lowest distance
        Advertiser lowest = new Advertiser("", new float[]{0.0f, 0.0f, 0.0f}, 1).add_measurement(999).add_measurement(999);
        Advertiser cur;

        List<String> keys = new ArrayList(ADVERTISERS.keySet());
        for (String e : keys) {
            cur = ADVERTISERS.get(e);
            lowest = cur.get_distance() < lowest.get_distance() ? cur : lowest;
        }
        return lowest;
    }

    /**
     * Gets called by pressing Button
     * 1th version should print postion via Log.d
     * Final version should set marker on map
     */
    private void get_position() {

        // gerneriere daten zur simulation
        simulationsdaten();

        s = null; //Distance
        beacon_variance = null; //Distance-Error;
        beacon_positions = null;

        List<String> keys = new ArrayList(ADVERTISERS.keySet());
        for (String e : keys) {
            Advertiser cur = ADVERTISERS.get(e);
            s.add( (double) cur.get_distance() );
            beacon_variance.add( cur.get_distance_error() );
            beacon_positions.add( cur.getPos() );
        }

        sx = vec_sub(advertiser_min_distance().getPos(), new float[]{0.0f, 0.0f, 0.0f});



        SimplexOptimizer optimizer = new SimplexOptimizer(1e-10, 1e-30);
        optimizer.setSimplex(new NelderMeadSimplex(new double[] { 0.2, 0.2 }));
        PointValuePair optimum
                = optimizer.optimize(100, new myfunc(), GoalType.MINIMIZE, new double[] { -3, 0 });
        double p = optimum.getPoint()[0];
    }

    public void simulationsdaten() {
        ADVERTISERS.clear();

        ADVERTISERS.put("00:07:80:52:64:e6", new Advertiser("00:07:80:52:64:e6", new float[]{2.5f, 5f, 1f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:7e:c3:68", new Advertiser("00:07:80:7e:c3:68", new float[]{1.0f, 0.5f, 0.5f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:7e:c3:7b", new Advertiser("00:07:80:7e:c3:7b", new float[]{5f, 2f, 1f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:68:1c:9c", new Advertiser("00:07:80:68:1c:9c", new float[]{0.5f, .5f, 2.2f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:68:28:29", new Advertiser("00:07:80:68:28:29", new float[]{6f, 5f, 1f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:68:28:67", new Advertiser("00:07:80:68:28:67", new float[]{3f, 3f, 0.5f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:79:1f:f1", new Advertiser("00:07:80:79:1f:f1", new float[]{6f, .5f, 1.8f}, 2).add_measurement(1).add_measurement(1) );
        ADVERTISERS.put("00:07:80:c0:ff:ee", new Advertiser("00:07:80:c0:ff:ee", new float[]{0.5f, 5f, 1f}, 2).add_measurement(1).add_measurement(1) );
    }


    public ArrayList<Double> getS() {
        return s;
    }

    public ArrayList<Float> getBeacon_variance() {
        return beacon_variance;
    }

    public ArrayList<float[]> getBeacon_positions() {
        return beacon_positions;
    }

    public float[] getSx() {
        return sx;
    }

    public Hashtable<String, Advertiser> getADVERTISERS() {
        return ADVERTISERS;
    }
}

